from Algorithms.algorithm import Algorithm
from niching_method.Kmeans import Kmeans
from niching_method.FS_niching import FS_niching
from localsearch.Localsearch import localsearch
from numpy import fabs, floor, asarray, exp, cos, pi, sqrt
import pandas as pd
import logging

logging.basicConfig()
logger = logging.getLogger('MMWOA')
logger.setLevel('INFO')


class MMWOA(Algorithm):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.K = kwargs.pop("K", 4)  # K value of K-means
        self.delta = kwargs.pop("delta", 1.0e-4)  # standard deviation of Gaussian distribution
        self.samplesize = kwargs.pop("samplesize", 4)  # Number of samples
        self.Glower = kwargs.pop("Glower", 2)  # Lower bound of cluster size in FS-clustering
        self.Gupper = kwargs.pop("Gupper", 20)  # Upper bound of cluster size in FS-clustering

    def encircling_prey(self, i, whale_pos, leader_pos, C, A):
        r"""
        :param i: Serial number of the individual
        :param whale_pos: Whale population
        :param leader_pos: Optimal whale position
        :param C: coefficient vector
        :param A: coefficient vector
        :return: Whale position after updating location
        """
        D = fabs(C * leader_pos - whale_pos.iloc[i, 2:])
        new_whale_pos = leader_pos - A * D
        return new_whale_pos

    def spiral_update_position(self, i, whale_pos, leader_pos, b, l):
        r"""
        :param i: Serial number of the individual
        :param whale_pos: Whale position
        :param leader_pos: Optimal whale position
        :param b: a constant
        :param l: a random vector
        :return: new whale position
        """
        distance = fabs(leader_pos - whale_pos.iloc[i, 2:])
        new_whale_pos = distance * exp(b * l) * cos(l * 2 * pi) + leader_pos
        return new_whale_pos

    def search_prey(self, i, whale_pos, C, A):
        r"""
        :param i: Serial number of the individual
        :param whale_pos: Whale position
        :param C: coefficient vector
        :param A: coefficient vector
        :return: new whale position
        """
        rand_leader_index = int(floor(len(whale_pos) * self.Rand.rand()))
        X_rand = whale_pos.iloc[rand_leader_index, 2:]
        D = fabs(C * X_rand - whale_pos.iloc[i, 2:])
        new_whale_pos = X_rand - A * D
        return new_whale_pos

    def update_position(self, i, whale_pos, leader_pos, a, a2):
        r"""
        :param i: Serial number of the individual
        :param whale_pos: Whale position
        :param leader_pos: Optimal whale position
        :param a: a number that decreased from 2 to 0
        :param a2: a number in [-1, 1]
        :return: new whale position
        """
        r1, r2 = self.Rand.rand(), self.Rand.rand()
        A = 2 * a * r1 - a
        C = 2 * r2
        b = 1
        l = (a2 - 1) * self.Rand.rand() + 1
        p = self.Rand.rand()
        if p < 0.5:
            if fabs(A) < 1:
                return self.encircling_prey(i, whale_pos, leader_pos, C, A)
            else:
                return self.search_prey(i, whale_pos, C, A)
        else:
            return self.spiral_update_position(i, whale_pos, leader_pos, b, l)

    def construct_new_whale(self, result, a, a2):
        r"""
        :param result: Clustered whale population
        :param a: a number that decreased from 2 to 0
        :param a2: a number in [-1, 1]
        :return: new whale population, fitness and cluster
        """
        new_whale = []
        new_whale_cluster = []
        new_whale_fit = []
        for name, group in result.groupby('c'):
            leader_pos = group['f'].idxmax()
            for i in range(len(group)):
                whale = self.update_position(i, result, result.iloc[leader_pos, 2:], a, a2)
                whale = self.boundary_handle(asarray(whale))
                new_whale.append(whale)
                new_whale_cluster.append(name)
                new_whale_fit.append(self.costfunction(whale))
        new_whale_cluster = pd.DataFrame(new_whale_cluster, columns=['c'])
        new_whale_fit = pd.DataFrame(new_whale_fit, columns=['f'])
        new_whale = pd.DataFrame(new_whale)
        return new_whale, new_whale_fit, new_whale_cluster

    def update_solutions(self, new_whale, new_whale_fit, new_whale_cluster, whale, whale_fit, whale_cluster):
        r"""
        :param new_whale: current new whale population
        :param new_whale_fit: the fitness of new whale population
        :param new_whale_cluster: cluster labels of new whale population
        :param whale: previous generation whale population
        :param whale_fit: the fitness of previous generation whale population
        :param whale_cluster: cluster labels of previous generation whale population
        :return: updated whale position
        """
        for i, row in new_whale.iterrows():
            distance_vector = []
            for j, r in whale.iterrows():
                distance_vector.append(sqrt((((row - r) ** 2).values).sum()))
            nearest_whale = distance_vector.index(min(distance_vector))
            if new_whale_fit.iloc[i]['f'] > whale_fit.iloc[nearest_whale]['f']:
                whale_fit.iloc[nearest_whale] = new_whale_fit.iloc[i]
                whale.iloc[nearest_whale] = row
                whale_cluster.iloc[nearest_whale] = new_whale_cluster.iloc[i]
        result = pd.concat([whale_cluster, whale_fit, whale], axis=1)
        return result

    def run(self):
        whale_pos = pd.DataFrame(self.initial_position())
        whale_fit = whale_pos.apply(self.costfunction, axis=1).astype('float')
        self.iter = 0
        seed = None

        while not self.stopping_criteria():
            self.iter += 1
            a = 2 - self.eval_counter * (2 / self.max_fes)
            a2 = -1 + self.eval_counter * (-1 / self.max_fes)

            # K-means clustering algorithm for niching
            whale_cluster, whale_fit, whale_pos, result = Kmeans(self.K, whale_pos, whale_fit).K_means()

            # Clustering algorithm with fixed cluster size bound for niching
            # nich_size = np.random.randint(self.Glower, self.Gupper)
            # whale_cluster, whale_fit, whale_pos, result = FS_niching(np.asarray(whale_pos), nich_size,
            #                                                          self.func).FS_Niching()

            new_whale, new_whale_fit, new_whale_cluster = self.construct_new_whale(result, a, a2)
            result = self.update_solutions(new_whale, new_whale_fit, new_whale_cluster, whale_pos, whale_fit,
                                           whale_cluster)
            result = localsearch(result, self.delta, self.samplesize, self.func).local_search()
            count, seed = how_many_goptima(pop=result.iloc[:, 2:].values, f=self.func, accuracy=1.0e-5)
            if count == self.func.get_no_goptima():
                break
        return seed


if __name__ == '__main__':
    from benchmarks.CEC2013.cec2013_function import *

    mmwoa = MMWOA(population=80, func=CEC2013_FUNCTION(1), seed=1)
    result = mmwoa.run()
    logger.info("{res}".format(res=result))
